import os
import sys
from sentence_transformers import SentenceTransformer
from transformers.trainer_callback import TrainerControl, TrainerState
from transformers.training_args import TrainingArguments
from nixietune.biencoder import BiencoderTrainer, BiencoderTrainingArguments
from nixietune.log import setup_logging
from transformers import HfArgumentParser, TrainerCallback
import logging
from nixietune import load_dataset_split, ModelArguments, DatasetArguments

setup_logging()


class EvaluateFirstStepCallback(TrainerCallback):
    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        if state.global_step == 0:
            control.should_evaluate = True

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_evaluate = True


def main():
    logger = logging.getLogger()
    parser = HfArgumentParser((ModelArguments, DatasetArguments, BiencoderTrainingArguments))
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        model_args, dataset_args, training_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, dataset_args, training_args = parser.parse_args_into_dataclasses()

    model = SentenceTransformer(model_args.model_name_or_path)
    train = load_dataset_split(dataset_args.train_dataset, split=dataset_args.train_split)
    test = load_dataset_split(dataset_args.eval_dataset, split=dataset_args.eval_split)

    logger.info(f"Training parameters: {training_args}")

    trainer = BiencoderTrainer(
        model=model,
        args=training_args,
        train_dataset=train,
        eval_dataset=test,
    )
    trainer.add_callback(EvaluateFirstStepCallback())
    trainer.train()

    model.save(path=training_args.output_dir)


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field


@dataclass
class DatasetArguments:
    train_dataset: str = field(metadata={"help": "Path to training dataset"})
    eval_dataset: str = field(metadata={"help": "Path to evaluation dataset"})
    train_split: str = field(default="train", metadata={"help": "name of the split for training"})
    eval_split: str = field(default="test", metadata={"help": "name of the split for evaluation"})


@dataclass
class ModelArguments:
    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )



from datasets import Features, Value
from transformers import PreTrainedTokenizerBase
from typing import Dict, List, Optional
from nixietune.format import Format
import numpy as np


class QueryDocLabelFormat(Format):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        query_prefix: Optional[str] = None,
        doc_prefix: Optional[str] = None,
        neg_count: int = 8,
    ) -> None:
        super().__init__(tokenizer, query_prefix, doc_prefix)
        self.neg_count = neg_count

    def tokenize(self, batch: Dict[str, List]) -> Dict[str, List]:
        token_cache = self.make_tokenized_cache(batch)

        features = []
        lengths = []
        labels = []
        for q, p, n in zip(batch.get("query", []), batch.get("positive", []), batch.get("negative", [])):
            query_tokens = token_cache[self.format_query(q)]
            for pos in p:
                doc = self.format_doc(pos)
                doc_tokens = token_cache[doc]
                features.append([query_tokens, doc_tokens])
                lengths.append(query_tokens["input_ids"].size + doc_tokens["input_ids"].size)
                labels.append(1.0)
            for neg in n:
                doc = self.format_doc(neg)
                doc_tokens = token_cache[doc]
                features.append([query_tokens, doc_tokens])
                lengths.append(query_tokens["input_ids"].size + doc_tokens["input_ids"].size)
                labels.append(0.0)
        result = {"features": features, "label": labels, "length": lengths}
        return result

    def schema(self, dtype: str) -> Features:
        return Features(
            {
                "features": [
                    {
                        "input_ids": [Value(dtype)],
                        "attention_mask": [Value("int8")],
                    }
                ],
                "label": Value("float"),
                "length": Value("int32"),
            }
        )



from abc import abstractmethod
from transformers import PreTrainedTokenizerBase
from typing import Dict, List, Optional
import numpy as np
from datasets import Features


class Format:
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        query_prefix: Optional[str],
        doc_prefix: Optional[str],
    ) -> None:
        self.tokenizer = tokenizer
        self.query_prefix = query_prefix
        self.doc_prefix = doc_prefix

    @abstractmethod
    def tokenize(self, batch: Dict[str, List]) -> Dict[str, List]:
        pass

    @abstractmethod
    def schema(self, dtype: str) -> Features:
        pass

    def make_tokenized_cache(self, batch: Dict[str, List]) -> Dict[str, Dict[str, np.ndarray]]:
        docs: List[str] = []
        for q, p, n in zip(batch["query"], batch.get("positive", []), batch.get("negative", [])):
            docs.append(self.format_query(q))
            for pos in p:
                docs.append(self.format_doc(pos))
            for neg in n:
                docs.append(self.format_doc(neg))
        tokens = self.tokenizer(docs, padding=False, truncation=True, return_tensors="np")
        token_cache: Dict[str, Dict] = {}
        for doc, input_ids, attention_mask in zip(docs, tokens.input_ids, tokens.attention_mask):
            token_cache[doc] = {
                "input_ids": input_ids.astype("int32"),
                "attention_mask": attention_mask.astype("int8"),
            }
        return token_cache

    def format_doc(self, doc: str) -> str:
        if self.doc_prefix is None:
            return doc
        else:
            return self.doc_prefix + doc

    def format_query(self, query: str) -> str:
        if self.query_prefix is None:
            return query
        else:
            return self.query_prefix + query



from sentence_transformers import SentenceTransformer, losses
from transformers import PreTrainedTokenizerBase
from typing import Optional
from nixietune.format import Format, QueryDocLabelFormat
from nixietune.target import Target
from torch import nn


class CosineSimilarityTarget(Target):
    def __init__(
        self,
        model: SentenceTransformer,
        tokenizer: PreTrainedTokenizerBase,
        query_prefix: Optional[str],
        doc_prefix: Optional[str],
    ) -> None:
        super().__init__(model, tokenizer, query_prefix, doc_prefix)

    def loss(self) -> nn.Module:
        return losses.CosineSimilarityLoss(self.model)

    def process(self) -> Format:
        return QueryDocLabelFormat(self.tokenizer, self.query_prefix, self.doc_prefix)



from typing import List, Dict
from torchmetrics.retrieval.ndcg import RetrievalNormalizedDCG
from torchmetrics.retrieval.average_precision import RetrievalMAP
from torchmetrics.retrieval.reciprocal_rank import RetrievalMRR
import torch
import logging
from transformers.trainer_utils import EvalPrediction
from sentence_transformers import util

logger = logging.getLogger()


class EvalMetrics:
    def __init__(self, metrics: List[str]) -> None:
        self.logger = logging.getLogger()
        self.metrics = metrics
        logger.info(f"Eval metrics: {metrics}")

    def compute(self, embeds: EvalPrediction) -> Dict[str, float]:
        query_embeds = torch.from_numpy(embeds.predictions[0]["sentence_embedding"]).cpu()
        doc_embeds = torch.from_numpy(embeds.predictions[1]["sentence_embedding"]).cpu()
        scores = util.pairwise_cos_sim(query_embeds, doc_embeds)
        _, indexes = torch.unique_consecutive(query_embeds, return_inverse=True, dim=0)
        targets = torch.from_numpy(embeds.label_ids)
        result = {}
        for metric_name in self.metrics:
            match metric_name.lower().split("@"):
                case ["ndcg"]:
                    metric = RetrievalNormalizedDCG()
                    result[metric_name] = metric(preds=scores, target=targets, indexes=indexes)
                case ["ndcg", k]:
                    metric = RetrievalNormalizedDCG(top_k=int(k))
                    result[metric_name] = metric(preds=scores, target=targets, indexes=indexes)
                case ["map"]:
                    result[metric_name] = RetrievalMAP()
                case ["map", k]:
                    result[metric_name] = RetrievalMAP(top_k=int(k))
                case ["mrr"]:
                    result[metric_name] = RetrievalMRR()
                case ["mrr", k]:
                    result[metric_name] = RetrievalMRR(top_k=int(k))
                case other:
                    logger.warn(f"Metric type {other} is not yet supported")

        return result



from nixietune.format.format import Format
from nixietune.format.query_doc_label import QueryDocLabelFormat
from nixietune.format.query_pos_negs import QueryPosNegsFormat
from nixietune.format.triplet import TripletFormat



from nixietune.target.target import Target
from nixietune.target.cosine import CosineSimilarityTarget
from nixietune.target.contrastive import ContrastiveTarget
from nixietune.target.infonce import InfoNCETarget
from nixietune.target.triplet import TripletTarget



from nixietune import ModelArguments, DatasetArguments, BiencoderTrainingArguments
from transformers import HfArgumentParser


def test_load_msmarco():
    parser = HfArgumentParser((ModelArguments, DatasetArguments, BiencoderTrainingArguments))
    model_args, dataset_args, training_args = parser.parse_json_file(json_file="examples/msmarco.json")
    assert model_args.model_name_or_path == "sentence-transformers/all-MiniLM-L6-v2"



from nixietune.biencoder import BiencoderTrainer, BiencoderTrainingArguments
from nixietune.format import Format, QueryDocLabelFormat, QueryPosNegsFormat, TripletFormat
from nixietune.arguments import ModelArguments, DatasetArguments
from datasets import DatasetDict, Dataset, load_dataset
import os
from huggingface_hub.utils import validate_repo_id
import logging

logger = logging.getLogger()


def load_dataset_split(path: str, split: str) -> Dataset:
    if os.path.exists(path):
        if os.path.isdir(path):
            return load_dataset("json", data_dir=path, split=split)
        else:
            return load_dataset("json", data_files={split: path}, split=split)
    else:
        validate_repo_id(path)
        return load_dataset(path, split=split)



from dataclasses import dataclass, field
from transformers import TrainingArguments
from typing import List, Dict, Any, Union, Tuple, Optional


@dataclass
class BiencoderTrainingArguments(TrainingArguments):
    seq_len: int = field(
        default=128,
        metadata={"help": "Max sequence length in tokens."},
    )

    target: str = field(
        default="infonce", metadata={"help": "Optimization target: cosine_similarity/contrastive/infonce"}
    )

    num_negatives: int = field(default=4, metadata={"help": "Number of negatives to use for InfoNCE/Triplet loss"})

    triplet_margin: float = field(default=5, metadata={"help": "Margin value for Triplet loss"})

    query_prefix: Optional[None] = field(
        default=None, metadata={"help": "Prefix for all queries. Used for asymmetrical models like E5."}
    )

    document_prefix: Optional[None] = field(
        default=None, metadata={"help": "Prefix for all documents. Used for asymmetrical models like E5."}
    )

    infonce_negative_mode: str = field(
        default="unpaired",
        metadata={
            "help": "Negative loss mode: paired (when loss is computed per query) / unpaired (when all negatives for all queries are taken at once)"
        },
    )

    infonce_temperature: float = field(default=0.05, metadata={"help": "Temperature for InfoNCE loss"})



from datasets import load_dataset, Features, Value, Sequence
from nixietune.log import setup_logging
import logging
from nixietune import load_dataset_split

setup_logging()
logger = logging.getLogger()


def test_dataset_loading():
    dataset = load_dataset("nixiesearch/ms-marco-dummy", split="train[:10%]")
    expected = Features(
        {
            "query": Value("string"),
            "positive": Sequence(Value("string")),
            "negative": Sequence(Value("string")),
        }
    )
    assert dataset.features == expected


def test_dataset_load_split_file():
    ds = load_dataset_split("tests/data/dummy.jsonl.gz", split="train")
    assert len(ds) == 10


def test_dataset_load_split_dir():
    ds = load_dataset_split("tests/data/", split="train")
    assert len(ds) == 10


def test_dataset_load_split_hf():
    ds = load_dataset_split("nixiesearch/ms-marco-dummy", split="train[:1%]")
    assert len(ds) == 10



from abc import abstractmethod
from torch import nn
from sentence_transformers import SentenceTransformer
from transformers import PreTrainedTokenizerBase
from typing import Optional
from nixietune.format import Format


class Target:
    def __init__(
        self,
        model: SentenceTransformer,
        tokenizer: PreTrainedTokenizerBase,
        query_prefix: Optional[str],
        doc_prefix: Optional[str],
    ) -> None:
        self.model = model
        self.tokenizer = tokenizer
        self.query_prefix = query_prefix
        self.doc_prefix = doc_prefix

    @abstractmethod
    def loss(self) -> nn.Module:
        pass

    @abstractmethod
    def process(self) -> Format:
        pass



from mteb import MTEB
import torch
from sentence_transformers import SentenceTransformer
from dataclasses import dataclass, field
from typing import Optional
from transformers import HfArgumentParser
import sys
import os


class SBERTModel:
    def __init__(
        self,
        name: str,
        query_prefix: Optional[str] = None,
        doc_prefix: Optional[str] = None,
    ):
        self.query_prefix = query_prefix
        self.doc_prefix = doc_prefix
        self.model = SentenceTransformer(name)
        self.model.to("cuda:0")
        self.model.eval()

    def encode_queries(self, queries, batch_size=32, **kwargs):
        if self.query_prefix is not None:
            processed = [f"{self.query_prefix}{q}" for q in queries]
        else:
            processed = queries
        return self.encode_impl(processed, batch_size)

    def encode_corpus(self, corpus, batch_size=32, **kwargs):
        if self.doc_prefix is not None:
            processed = [f"{self.query_prefix}{q}" for q in corpus]
        else:
            processed = corpus
        return self.encode_impl(processed, batch_size)

    @torch.no_grad()
    def encode_impl(self, input_texts, batch_size):
        return self.model.encode(input_texts, normalize_embeddings=True, batch_size=batch_size)


@dataclass
class EvalArguments:
    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    result_path: str = field(metadata={"help": "Output path for results"})
    batch_size: int = field(default=256, metadata={"help": "Batch size for model inference"})


def main():
    parser = HfArgumentParser((EvalArguments))
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        (eval_args,) = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        (eval_args,) = parser.parse_args_into_dataclasses()

    model = SBERTModel(eval_args.model_name_or_path)
    evaluation = MTEB(tasks=["FiQA2018", "SciFact", "NFCorpus"])
    evaluation.run(model, output_folder=eval_args.result_path, batch_size=eval_args.batch_size, verbosity=2)


if __name__ == "__main__":
    main()



from datasets import load_dataset
from nixietune.log import setup_logging
from nixietune.format import TripletFormat, QueryDocLabelFormat, QueryPosNegsFormat
from transformers import AutoTokenizer
import pytest


@pytest.fixture
def dataset():
    return load_dataset("nixiesearch/ms-marco-dummy", split="train[:10%]")


@pytest.fixture
def tokenizer():
    return AutoTokenizer.from_pretrained("sentence-transformers/all-MiniLM-L6-v2")


def test_triplet_format(dataset, tokenizer):
    parser = TripletFormat(tokenizer)
    _test_format(dataset, parser)


def test_querydoclabel_format(dataset, tokenizer):
    parser = QueryDocLabelFormat(tokenizer)
    _test_format(dataset, parser)


def test_queryposnegs_format(dataset, tokenizer):
    parser = QueryPosNegsFormat(tokenizer)
    _test_format(dataset, parser)


def _test_format(dataset, format):
    schema = format.schema("uint16")
    processed = dataset.map(
        function=format.tokenize,
        batched=True,
        remove_columns=["query", "positive", "negative"],
        features=schema,
    )
    assert processed.features == schema



from transformers import PreTrainedModel
from sentence_transformers import SentenceTransformer
from transformers import Trainer
from typing import List, Dict, Any, Union, Tuple
import torch
from torch import nn

from datasets import Dataset
from nixietune.metrics import EvalMetrics
from nixietune.target import CosineSimilarityTarget, ContrastiveTarget, InfoNCETarget, TripletTarget
from nixietune.biencoder import BiencoderTrainingArguments
import logging
from transformers.tokenization_utils_base import BatchEncoding
from tqdm import tqdm
import numpy as np
from itertools import islice

logger = logging.getLogger()


class BiencoderModel(PreTrainedModel):
    supports_gradient_checkpointing = True

    def __init__(self, model: nn.Sequential):
        super().__init__(model[0].auto_model.config)
        self.model = model

    def forward(self, tensor, return_loss: bool = False):
        return self.model.forward(tensor)


class BiencoderTrainer(Trainer):
    def __init__(
        self,
        model: SentenceTransformer,
        args: BiencoderTrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        eval_metrics: List[str] = ["ndcg@10"],
        **kwargs,
    ) -> None:
        self.eval_metrics = EvalMetrics(eval_metrics)
        tokenizer = model.tokenizer
        tokenizer.model_max_length = args.seq_len
        tokenizer.deprecation_warnings["Asking-to-pad-a-fast-tokenizer"] = True
        match args.target:
            case "cosine_similarity":
                self.target = CosineSimilarityTarget(model, tokenizer, args.query_prefix, args.document_prefix)
            case "contrastive":
                self.target = ContrastiveTarget(model, tokenizer, args.query_prefix, args.document_prefix)
            case "infonce":
                self.target = InfoNCETarget(
                    model, tokenizer, args.num_negatives, args.query_prefix, args.document_prefix
                )
            case "triplet":
                self.target = TripletTarget(
                    model, tokenizer, args.query_prefix, args.document_prefix, args.num_negatives, args.triplet_margin
                )
        self.eval_target = CosineSimilarityTarget(model, tokenizer, args.query_prefix, args.document_prefix)
        self.processor = self.target.process()
        self.eval_processor = self.eval_target.process()
        self.loss = self.target.loss()
        self.loss.to(model.device)
        self.eval_loss = self.eval_target.loss()
        self.eval_loss.to(model.device)
        args.label_names = ["label"]
        args.gradient_checkpointing_kwargs = {"use_reentrant": False}
        args.remove_unused_columns = False
        # self.print_raw_stats(train_dataset)
        if len(tokenizer) < 65536:
            dtype = "uint16"
        else:
            dtype = "uint32"
        logger.info(f"Tokenizer vocab_size={len(tokenizer)}, using {dtype} for token encoding on disk cache")
        train_processed = train_dataset.map(
            self.processor.tokenize,
            batched=True,
            batch_size=128,
            num_proc=args.dataloader_num_workers,
            remove_columns=["query", "positive", "negative"],
            desc="Tokenizing train dataset",
            features=self.processor.schema(dtype),
        )
        # self.print_tokenized_stats(train_processed)
        eval_processed = eval_dataset.map(
            self.eval_processor.tokenize,
            batched=True,
            batch_size=128,
            num_proc=args.dataloader_num_workers,
            remove_columns=["query", "positive", "negative"],
            desc="Tokenizing test dataset",
            features=self.eval_processor.schema(dtype),
        )
        bi_model = BiencoderModel(model)
        bi_model.warnings_issued["estimate_tokens"] = True
        super().__init__(
            args=args,
            model=bi_model,
            compute_metrics=self.eval_metrics.compute,
            preprocess_logits_for_metrics=self.move_to_cpu,
            train_dataset=train_processed,
            eval_dataset=eval_processed,
            data_collator=self.collate,
            tokenizer=tokenizer,
            **kwargs,
        )

    def compute_loss(
        self,
        model: SentenceTransformer,
        inputs: Dict[str, Dict[str, torch.Tensor]],
        return_outputs: bool = False,
    ) -> Union[Tuple[torch.Tensor, Dict[str, torch.Tensor]], torch.Tensor]:
        features = []
        for column in inputs["features"]:
            features.append({"input_ids": column.input_ids, "attention_mask": column.attention_mask})
        if return_outputs is True:
            target_loss = self.eval_loss
        else:
            target_loss = self.loss
        if "label" in inputs:
            loss = target_loss(features, inputs["label"])
        else:
            loss = target_loss(features, torch.Tensor())
        if return_outputs:
            output = [torch.Tensor()] + features
            return loss, output
        return loss

    def move_to_cpu(self, logits: List[Dict[str, torch.Tensor]], labels: torch.Tensor) -> List[torch.Tensor]:
        result = []
        for f in logits:
            se = f["sentence_embedding"]
            result.append({"sentence_embedding": se.clone()})
        return result

    def collate(self, items: List[Dict[str, Dict]]) -> Dict[str, Dict[str, torch.Tensor]]:
        features = []
        for item in items:
            for index, feature in enumerate(item["features"]):
                if index >= len(features):
                    features.extend([[]] * (index + 1 - len(features)))
                features[index].append(feature)

        padded_features = [self.pad(f) for f in features]
        result = {}
        result["features"] = padded_features
        if "label" in items[0]:
            result["label"] = torch.tensor([item["label"] for item in items])
        else:
            result["return_loss"] = True
        return result

    def pad(self, docs: List[Dict[str, Any]]) -> BatchEncoding:
        batch = BatchEncoding(
            data={
                "input_ids": [f["input_ids"] for f in docs],
                "attention_mask": [f["attention_mask"] for f in docs],
            }
        )
        return self.tokenizer.pad(batch, padding="longest", pad_to_multiple_of=8, return_tensors="pt")

    def print_raw_stats(self, dataset: Dataset, samples: int = 5000) -> None:
        positives_per_query = []
        negatives_per_query = []
        for row in tqdm(islice(dataset, samples), desc="Collecting raw stats", total=samples):
            pos = len(row["positive"])
            positives_per_query.append(pos)
            neg = len(row["negative"])
            negatives_per_query.append(neg)
        ppq = np.percentile(positives_per_query, [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
        npq = np.percentile(negatives_per_query, [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
        logger.info(f"Negatives per query: {npq}")
        logger.info(f"Positives per query: {ppq}")

    def print_tokenized_stats(self, dataset: Dataset, samples: int = 5000) -> None:
        query_tokens = []
        doc_tokens = []
        for row in tqdm(islice(dataset, samples), desc="Collecting token stats", total=samples):
            features = row["features"]
            query, positive, *docs = features
            query_tokens.append(len(query["input_ids"]))
            doc_tokens.append(len(positive["input_ids"]))
            [doc_tokens.append(len(neg["input_ids"])) for neg in docs]
        qtp = np.percentile(query_tokens, [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100], method="nearest")
        dtp = np.percentile(doc_tokens, [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100], method="nearest")
        logger.info(f"Query tokens:    {qtp}")
        logger.info(f"Document tokens: {dtp}")



from sentence_transformers import SentenceTransformer, losses
from transformers import PreTrainedTokenizerBase
from typing import Optional
from nixietune.format import Format, TripletFormat
from nixietune.target import Target
from torch import nn


class TripletTarget(Target):
    def __init__(
        self,
        model: SentenceTransformer,
        tokenizer: PreTrainedTokenizerBase,
        query_prefix: Optional[str],
        doc_prefix: Optional[str],
        num_negs: int,
        margin: float,
    ) -> None:
        super().__init__(model, tokenizer, query_prefix, doc_prefix)
        self.num_negs = num_negs
        self.margin = margin

    def loss(self) -> nn.Module:
        return losses.TripletLoss(
            self.model, distance_metric=losses.TripletDistanceMetric.COSINE, triplet_margin=self.margin
        )

    def process(self) -> Format:
        return TripletFormat(self.tokenizer, self.query_prefix, self.doc_prefix, neg_count=self.num_negs)



from sentence_transformers import SentenceTransformer, losses
from transformers import PreTrainedTokenizerBase
from typing import Optional, Iterable, Dict
from nixietune.format import Format, QueryPosNegsFormat
from nixietune.target import Target
from torch import nn, Tensor
from info_nce import info_nce
import torch


class InfoNCELoss(nn.Module):
    def __init__(self, model: SentenceTransformer, temperature=0.05, reduction="mean", negative_mode="unpaired"):
        super().__init__()
        self.temperature = temperature
        self.reduction = reduction
        self.negative_mode = negative_mode
        self.model = model

    def forward(self, sentence_features: Iterable[Dict[str, Tensor]], labels: Tensor):
        reps = [self.model(sentence_feature)["sentence_embedding"] for sentence_feature in sentence_features]
        queries = reps[0]
        positives = reps[1]
        if self.negative_mode == "paired":
            if len(reps) >= 2:
                negatives = torch.stack(reps[2:], dim=1)
            else:
                raise ValueError("Cannot use paired InfoNCE loss with no negatives")
            return info_nce(
                queries,
                positives,
                negatives,
                temperature=self.temperature,
                reduction=self.reduction,
                negative_mode=self.negative_mode,
            )
        elif self.negative_mode == "unpaired":
            if len(reps) >= 2:
                negatives = torch.cat(reps[2:])
            else:
                negatives = None
            return info_nce(
                queries,
                positives,
                negatives,
                temperature=self.temperature,
                reduction=self.reduction,
                negative_mode=self.negative_mode,
            )


class InfoNCETarget(Target):
    def __init__(
        self,
        model: SentenceTransformer,
        tokenizer: PreTrainedTokenizerBase,
        num_negs: int,
        query_prefix: Optional[str],
        doc_prefix: Optional[str],
        temperature: float,
        negative_mode: str,
    ) -> None:
        super().__init__(model, tokenizer, query_prefix, doc_prefix)
        self.num_negs = num_negs
        self.temperature = temperature
        self.negative_mode = negative_mode

    def loss(self) -> nn.Module:
        # return losses.MultipleNegativesRankingLoss(self.model)
        return InfoNCELoss(self.model, negative_mode=self.negative_mode, temperature=self.temperature)

    def process(self) -> Format:
        return QueryPosNegsFormat(self.tokenizer, self.query_prefix, self.doc_prefix, self.num_negs)



import logging
import transformers


def setup_logging():
    log_level = logging.INFO
    log_format = logging.Formatter("[%(asctime)s %(levelname)s] %(message)s")
    logger = logging.getLogger()
    logger.setLevel(log_level)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_format)
    logger.handlers = [console_handler]
    # datasets.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()
